
#include "stm32f10x.h"
#include "stdlib.h"
#include "stdbool.h"
#include "stdarg.h"
#include "math.h"

//
// - DCFMeter - misst Resonanzfrequenz, Bandbreite und Gesamtempfindlichkeit
//
//   Hilfsprogramm um Parameter vom DCF77-Schwingkreis inkl. Verstrker-
//   elektronik zu messen.
//
// - Schaltplan / Aufbau
//   - Sendespule (s.u. "Coil_d_mm" und "Coil_N_1") mit PB6 und PB7 verbinden
//   - von VCC je ein Widerstand (s.u. "Coil_Resistor_Ohm") an PB6 und PB7
//   - die Leitungen an PB6 sollten dicht zu den jeweiligen Leitungen an PB7
//     gefhrt werden
//   - die VCC- und die GND-Leitung (wenn nicht entstrt enthalten diese
//     quasi auch das Sendesignal) sollten dicht beieinander und fern vom
//     Schwingkreis liegen
//   - den zu testenden Schwingkreis im Zentrum der Sendespule aufstellen
//   - Verstrker normal anschlieen
//   - mind. 1m Abstand zu Strquellen (Rechner, ...)
//   - mind. 0.5m Abstand (hngt von der Sendespulengre ab) zu greren
//     Metallteilen (Stahlbeton, Bleche, ...)
// - Achtung!
//   Bei Sendespulenwiderstnden kleiner 10000Ohm ("Coil_Resistor_Ohm") knnte
//   der Controller beschdigt werden!
// - Messergebnis
//   Das Ergebnis wird an USART1 (9600/8-N-1 an GPIO_PA9) ausgegeben.
// - Bedienungsanleitung
//   Der ber USART angezeigte Wert fr UssADC sollte zwischen 100mV (wegen
//   Strsignalen) und 1000mV (wegen Verstrkerverzerrungen) liegen. Das kann
//   durch nderung der Spulenparameter (Durchmesser, Windungszahl und
//   Widerstand) erreicht werden. Wenn die Spulenparameter von den Quelltext-
//   werten abweichen, dann kann man sich fr "Uv/E" den entsprechenden
//   Korrekturfaktor hinzudenken ("k = (d/d0)*(N0/N)*(R/R0)").
//   Je weiter die Empfangsspule aus den Zentrum der Sendespule ragt, desto
//   hher wird der Messfehler fr "Uv/E". Bei einer Empfangsspulenlnge von
//   10% des Sendespulendurchmessers ist "Uv/E" um 2% zu niedrig und bei 20%
//   um 7%.
//   Es kann (geringe Abstnde, lange Schwingkreiszuleitung) sinnvoll sein
//   die Messung mit umgekehrter Spulenpolaritt zu wiederholden. Je mehr
//   sich die Werte ndern, desto hher ist der vom Verstrker oder dessen
//   Zuleitung eingestreute EM-Feldanteil (durch Stromflsse oder den
//   Ausgangsspannungshub). Bis etwa 20% Abweichnung ist eigentlich der
//   Mittelwert noch brauchbar, andernfalls hilft ein besserer Aufbau.
//   Nicht vergessen: Metallteile (z.B. Krokodilklemmen) in unmittelbarer
//   Spulennhe knnen Schwingkreisparameter deutlich beeinflussen.
// - Technische Daten
//   - 23kHz - 142kHz Resonanzfrequenz
//   - x*1% Genauigkeit
//
// Copyright: Public Domain
//
// C.B., 2021
//

// "OPTS" : damit beim Debug-Kompilat erstens "fMHz" reicht und zweitens das Programm noch im Ram (statt Flash) passt
#define OPTS __attribute__((optimize("-Os")))

#define my_assert(Condition, InfoVal)  if(!(Condition) && _Error==_ErrOk) _Error=InfoVal;

enum
{
  // Parameter der Sendespule
  Coil_d_mm           =   250,  // Durchmesser der Sendespule
  Coil_N_1            =     1,  // Windungszahl der Sendespule
  Coil_Resistor_Ohm   = 33000,  // Pullup-Widerstand (10k bis 100k) der Sendespule (davon mssen zwei Stck verbaut sein)

  STM32_VCC_mV        = 3300,

  fMHz                = 64, // fest
  PeriodsPerSample    = 10, // so wie in DCF77.h
  ADCSamplesPerPeriod = 12, // so wie in DCF77.h

  ADCRange            = 4096,  // STM32-Konstante

  SkipPercentage      = 10,  // wieviel Prozent der Messzeit sollen wegen des Einschwingens ignoriert werden

  ScanFreqCnt         = 200, // Anzahl der Messwerte zwischen MinFrequency und MaxFrequency
  ScanAvrCnt          =  10, // Anzahl Messwerte zur Berechnung eines Amplitudenmittelwertes
  Scan_Cycle_ms       = 200, // Scan-Zeit (Gesamtzeit fr alle Messpunkte)

  IterationCnt        =  5, // -> Iterationszeit (=(2^IterationCnt)*(3*ScanAvrCnt*Scan_Cycle_ms/ScanFreqCnt))

  // berechnet
  AdcBufCnt    = 2*PeriodsPerSample*ADCSamplesPerPeriod,
  MinFrequency = fMHz*1000*1000/ADCSamplesPerPeriod/250, // wegen uint8_t in TimArr
  MaxFrequency = fMHz*1000*1000/ADCSamplesPerPeriod /32, // wegen Mindest-ADC-Takte
  fStep        = (int32_t)(1/(exp(1.0/ScanFreqCnt*log(MaxFrequency/MinFrequency))-1)),

};

enum TDCF77_Error
{
  _ErrOk                  =  0, // kein Fehler
  _Err_fMHz               =  1,
  _ErrDCT                 =  2, // fr "ADCSamplesPerPeriod" gibt es keinen Source-Code
  _ErrTimedAdc            =  3, // fCPU zu gross bzw. ADCSamplesPerPeriod zu klein bzw. TIM2 braucht Prescaler
};

typedef struct
{
  int16_t ValArr[AdcBufCnt]; // Puffer ADC-Werte
  uint8_t TimArr[AdcBufCnt]; // ADC-Sample-Zeitpunktabstnde
  int32_t w;
  int32_t n;
  int32_t a0;
  int32_t a1;
  int32_t d_ppm_L10;
} TTimedAdc;

static int _Error = _ErrOk;

//################################################### TimedAdc_Init
static int32_t* OPTS TimedAdc_Init(TTimedAdc* s, int32_t ppm, int fCarrier)
{
  // - Quarzabweichung ausgleichbar
  // - Grobabgleich per ppm-Parameter in 1 ppm Schritten
  //   (positiver Wert: Quarz schwingt schneller als Nominalfrequenz)
  // - Feinabgleich per TimedAdc_Tweak() in 1/256 ppm Schritten
  // - verursacht ca. +-2 Trgerphasenrauschen bei fCPU=16MHz und fCarrier=77.5kHz (=77.5kHz/16MHz*360)
  // - der DMA darf weiterlaufen
  int32_t fADC = fCarrier*ADCSamplesPerPeriod;
  int n0 = fMHz*(1000*1000+ppm)/fADC; // STM32F030 DIV?
  int a0 = fMHz*(1000*1000+ppm)-fADC*n0;
  my_assert(n0<=254, _ErrTimedAdc);
  int a1 = fADC;
  int w = -a1;
  int ns = 0;
  for(int i=0; i<AdcBufCnt; i++)
  {
    int n = n0;
    w += a0;
    if(w>0)
    {
      n = n0+1;
      w -= a1;
    }
    s->TimArr[i] = n-1;
    ns += n;
  }
  s->n  = n0;
  s->a0 = fMHz*(1000*1000+ppm)*AdcBufCnt-fADC*ns;
  s->a1 = fADC;
  s->w  = -s->a1;
  return &s->d_ppm_L10; // (darf von auen manipuliert werden)
}

//################################################### TimedAdc_Tweak
static void OPTS TimedAdc_Tweak(TTimedAdc* s)
{
  // je TimedAdc.TimArr[]-DMA-Durchlauf _genau_einmal_ aufrufen:
  // Der Feinabgleichsparameter "d_ppm_L10" darf jederzeit von auen
  // gendert werden.
  int n = s->n; // Standard (kein extra CPU-Takt)
  s->w += s->a0;
  if(s->w>0)
  {
    n = s->n+1; // Sonderfall: ein CPU-Takt extra
    s->w -= s->a1;
  }
  s->TimArr[0] = n-1;
  // neuen Korrekturwert anwenden?
  if(s->d_ppm_L10!=0)
  {
    int a0 = s->a0+(s->d_ppm_L10*AdcBufCnt*fMHz)/1024;
    s->d_ppm_L10 = 0;
    int d = 0;
    if(a0<0) d=+1;
    if(a0>=s->a1) d=-1;
    s->n -= d;
    s->a0 = a0+d*s->a1;
  }
}

//################################################### DoDCT
static void OPTS DoDCT(int16_t* Arr, int32_t PeriodCnt, int32_t* ValC, int32_t* ValS)
{
  // berlagerung mit Amplitude von 256
  my_assert(ADCSamplesPerPeriod>=3 && ADCSamplesPerPeriod<=12, _ErrDCT);
  int32_t SumC = *ValC;
  int32_t SumS = *ValS;
  do
  {
    int32_t v;
    // (Code wurde automatisiert erstellt)
    if(ADCSamplesPerPeriod==3)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==4)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 2
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS -= v*256;
    }
    if(ADCSamplesPerPeriod==5)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 2
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 3
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 4
      SumC += v* 79;
      SumS -= v*243;
    }
    if(ADCSamplesPerPeriod==6)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 2
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 5
      SumC += v*128;
      SumS -= v*221;
    }
    if(ADCSamplesPerPeriod==7)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*159;
      SumS += v*200;
      v = *Arr++;    // 2
      SumC -= v* 57;
      SumS += v*249;
      v = *Arr++;    // 3
      SumC -= v*230;
      SumS += v*111;
      v = *Arr++;    // 4
      SumC -= v*230;
      SumS -= v*111;
      v = *Arr++;    // 5
      SumC -= v* 57;
      SumS -= v*249;
      v = *Arr++;    // 6
      SumC += v*159;
      SumS -= v*200;
    }
    if(ADCSamplesPerPeriod==8)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*181;
      SumS += v*181;
      v = *Arr++;    // 2
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 3
      SumC -= v*181;
      SumS += v*181;
      v = *Arr++;    // 4
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 5
      SumC -= v*181;
      SumS -= v*181;
      v = *Arr++;    // 6
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 7
      SumC += v*181;
      SumS -= v*181;
    }
    if(ADCSamplesPerPeriod==9)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*196;
      SumS += v*164;
      v = *Arr++;    // 2
      SumC += v* 44;
      SumS += v*252;
      v = *Arr++;    // 3
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 4
      SumC -= v*240;
      SumS += v* 87;
      v = *Arr++;    // 5
      SumC -= v*240;
      SumS -= v* 87;
      v = *Arr++;    // 6
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 7
      SumC += v* 44;
      SumS -= v*252;
      v = *Arr++;    // 8
      SumC += v*196;
      SumS -= v*164;
    }
    if(ADCSamplesPerPeriod==10)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*207;
      SumS += v*150;
      v = *Arr++;    // 2
      SumC += v* 79;
      SumS += v*243;
      v = *Arr++;    // 3
      SumC -= v* 79;
      SumS += v*243;
      v = *Arr++;    // 4
      SumC -= v*207;
      SumS += v*150;
      v = *Arr++;    // 5
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 6
      SumC -= v*207;
      SumS -= v*150;
      v = *Arr++;    // 7
      SumC -= v* 79;
      SumS -= v*243;
      v = *Arr++;    // 8
      SumC += v* 79;
      SumS -= v*243;
      v = *Arr++;    // 9
      SumC += v*207;
      SumS -= v*150;
    }
    if(ADCSamplesPerPeriod==11)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*215;
      SumS += v*138;
      v = *Arr++;    // 2
      SumC += v*106;
      SumS += v*232;
      v = *Arr++;    // 3
      SumC -= v* 37;
      SumS += v*253;
      v = *Arr++;    // 4
      SumC -= v*167;
      SumS += v*193;
      v = *Arr++;    // 5
      SumC -= v*245;
      SumS += v* 72;
      v = *Arr++;    // 6
      SumC -= v*245;
      SumS -= v* 72;
      v = *Arr++;    // 7
      SumC -= v*167;
      SumS -= v*193;
      v = *Arr++;    // 8
      SumC -= v* 37;
      SumS -= v*253;
      v = *Arr++;    // 9
      SumC += v*106;
      SumS -= v*232;
      v = *Arr++;    // 10
      SumC += v*215;
      SumS -= v*138;
    }
    if(ADCSamplesPerPeriod==12)
    {
      v = *Arr++;    // 0
      SumC += v*256;
      SumS += v*  0;
      v = *Arr++;    // 1
      SumC += v*221;
      SumS += v*127;
      v = *Arr++;    // 2
      SumC += v*128;
      SumS += v*221;
      v = *Arr++;    // 3
      SumC += v*  0;
      SumS += v*256;
      v = *Arr++;    // 4
      SumC -= v*128;
      SumS += v*221;
      v = *Arr++;    // 5
      SumC -= v*221;
      SumS += v*127;
      v = *Arr++;    // 6
      SumC -= v*256;
      SumS += v*  0;
      v = *Arr++;    // 7
      SumC -= v*221;
      SumS -= v*127;
      v = *Arr++;    // 8
      SumC -= v*128;
      SumS -= v*221;
      v = *Arr++;    // 9
      SumC += v*  0;
      SumS -= v*256;
      v = *Arr++;    // 10
      SumC += v*128;
      SumS -= v*221;
      v = *Arr++;    // 11
      SumC += v*221;
      SumS -= v*127;
    }
    if(ADCSamplesPerPeriod==8 && 1==2)
    {
      enum
      {
        ScaleSin00 =   0,
        ScaleSin45 = 256,
        ScaleSin90 = 362  // =ScaleSin45/sin(45)
      };
      int v;
      //                    // #   cos   sin
      v = *Arr++;
      SumC += v*ScaleSin90; // 0  +100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC += v*ScaleSin45; // 1   +71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 2     0  +100
      SumS += v*ScaleSin90;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 3   -71   +71
      SumS += v*ScaleSin45;
      v = *Arr++;
      SumC -= v*ScaleSin90; // 4  -100     0
      SumS += v*ScaleSin00;
      v = *Arr++;
      SumC -= v*ScaleSin45; // 5   -71   -71
      SumS -= v*ScaleSin45;
      v = *Arr++;
      SumC += v*ScaleSin00; // 6     0  -100
      SumS -= v*ScaleSin90;
      v = *Arr++;
      SumC += v*ScaleSin45; // 7   +71   -71
      SumS -= v*ScaleSin45;
    }
  } while(--PeriodCnt>0);
  *ValC = SumC;
  *ValS = SumS;
}

//################################################### uprintf
static void uprintf_chrout(char Char);
static int  uprintf_GetNum(const char** Fmt);
static void uprintf(const char* Fmt, ...)
{
  // "mini-usart-printf"
  // (Einweg - nicht wiederverwenden)
  va_list va;
  va_start(va, Fmt);
  for(; *Fmt!=0; Fmt++)
  {
    if(*Fmt=='%')
      {
        bool Zero = *(++Fmt)=='0';
        int  n0 = uprintf_GetNum(&Fmt);
        int  n1 = Zero ? n0 : 0;
        if(Zero) n1=n0;
        if(*Fmt=='.')
        {
          Fmt++;
          n1 = uprintf_GetNum(&Fmt)+1;
        }
        switch(*Fmt)
        {
          case 'c': {
                      uprintf_chrout(va_arg(va, int32_t));
                      break;
                    }
          /*case 's': {
                      const char* s = va_arg(va, const char*);
                      if(n0!=0)
                      {
                        int n2 = 0;
                        for(; s[n2]!=0; n2++) { };
                        for(; n2<n0; n2++) uprintf_chrout(' ');
                      }
                      while(*s!=0) uprintf_chrout(*s++);
                      break;
                    }*/
          case 'i': {
                      int v = va_arg(va, int32_t);
                      int cv = 10;
                      int n2 = 1;
                      for(; abs(v)>=cv; n2++) cv*=10;
                      int nn = n1>n2 ? n1 : n2;
                      int nSpace = n0-nn;
                      if(v<0) nSpace--;
                      for(; nSpace>0; nSpace--) uprintf_chrout(' ');
                      if(v<0)
                      {
                        uprintf_chrout('-');
                        v = abs(v);
                      }
                      for(int i=0; i<nn; i++)
                      {
                        if(i>=n1-n2) cv/=10;
                        int n = v/cv;
                        uprintf_chrout(n+48);
                        v -= n*cv;
                        if(!Zero && i>=n2-n1)
                        {
                          uprintf_chrout('.');
                          Zero = true;
                        }

                      }
                      break;
                    }
        }
      }
    else
      uprintf_chrout(*Fmt);
  }
  va_end(va);
}
static void uprintf_chrout(char Char)
{
  // TXD-Ausgang ist GPIO_PA9
  const int Baud = 9600;
  if((USART1->CR1&USART_CR1_UE)==0)
  {
    RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
    GPIOA->CRH = (GPIOA->CRH&0xffffff0f)|0x000000a0;
    RCC->APB2ENR |= RCC_APB2ENR_USART1EN;
    USART1->CR1 = USART_CR1_UE;
    int divval = (fMHz/1*1000000+Baud/2)/Baud; // 16..65535
    USART1->BRR = divval;
    USART1->CR1 |= USART_CR1_TE;
  }
  USART1->DR = Char;
  while((USART1->SR&USART_SR_TXE)==0)
  {
    // __WFI();
  }
}
static int uprintf_GetNum(const char** Fmt)
{
  int v = 0;
  while(**Fmt>='0' && **Fmt<='9')
  {
    v = v*10+(*(*Fmt)++-'0');
  }
  return v;
}

//############################################################## int64_sqrt
static int32_t OPTS int64_sqrt(int64_t x)
{
  // https://en.wikipedia.org/wiki/Integer_square_root
  int shift = 2;
  while((x>>shift)!=0) shift+=2;
  int y = 0;
  for(; shift>=0; shift-=2)
  {
    y <<= 1;
    int64_t v = y+1;
    if(v*v<=(x>>shift)) y=(int)v;
  }
  return y;
}

//############################################################## Measure
static int32_t OPTS Measure(TTimedAdc* TimedAdc,
                            int32_t    Frequency,
                            int32_t    NumPeriods)
{
  int nSkip = SkipPercentage*NumPeriods/100;
  if(nSkip<PeriodsPerSample) nSkip=PeriodsPerSample;
  int nPeriods = NumPeriods-nSkip;
  if(nPeriods<PeriodsPerSample) nPeriods=PeriodsPerSample;
  int64_t SumC = 0;
  int64_t SumS = 0;
  int32_t SumN = -nSkip;
  if(Frequency<MinFrequency || Frequency>MaxFrequency) return 1;
  TimedAdc_Init(TimedAdc, 0, Frequency);
  while(true)
  {
    int Ofs = -1;
    {
      if((DMA1->ISR&DMA_ISR_HTIF1)!=0)
      {
        Ofs = 0*PeriodsPerSample*ADCSamplesPerPeriod;
        DMA1->IFCR = DMA_IFCR_CHTIF1;
      }
      if((DMA1->ISR&DMA_ISR_TCIF1)!=0)
      {
        Ofs = 1*PeriodsPerSample*ADCSamplesPerPeriod;
        DMA1->IFCR = DMA_IFCR_CTCIF1;
      }
    }
    if(Ofs>=0)
    {
      int32_t ValC = 0;
      int32_t ValS = 0;
      DoDCT(TimedAdc->ValArr+Ofs, PeriodsPerSample, &ValC, &ValS);
      if(SumN>=0)
      {
        SumC += ValC;
        SumS += ValS;
      }
      SumN += PeriodsPerSample;
      if(SumN>=nPeriods) break;
    }
    // Feinabgleich von fADC
    if((DMA1->ISR&DMA_ISR_TCIF2)!=0 && DMA1_Channel2->CNDTR<AdcBufCnt-1)
    {
      DMA1->IFCR = DMA_IFCR_CTCIF2;
      TimedAdc_Tweak(TimedAdc);
    }
  }
  SumC = SumC/SumN/ADCSamplesPerPeriod;
  SumS = SumS/SumN/ADCSamplesPerPeriod;
  int Amp = int64_sqrt(SumC*SumC+SumS*SumS);
  return Amp<1 ? 1 : Amp;
}

//################################################### main
int main(void)
{
  TTimedAdc TimedAdc;
  int32_t AmpArr[ScanFreqCnt];

  // HSE auf 64 MHz (Q&D)
  {
    my_assert(fMHz==64, _Err_fMHz);
    RCC->CFGR = 0;
    RCC->CR &= ~(RCC_CR_HSEON|RCC_CR_PLLON);
    RCC->CR |= RCC_CR_HSEON;
    while((RCC->CR|RCC_CR_HSERDY)==0);
    RCC->CFGR |= RCC_CFGR_PLLSRC;
    RCC->CFGR |= RCC_CFGR_SW_0*1;
    while((RCC->CFGR&RCC_CFGR_SWS)!=RCC_CFGR_SWS_0*1);
    RCC->CFGR |= RCC_CFGR_PLLMULL8; // 8*(8MHz/1)
    FLASH->ACR = FLASH_ACR_LATENCY_1*2;
    RCC->CR |= RCC_CR_PLLON;
    RCC->CFGR = (RCC->CFGR&~RCC_CFGR_SW)|(RCC_CFGR_SW_0*2);
  }

  // Init GPIO
  // - alle unbelegten GPIO-Pins auf Pullup damit STM32-Eingangslogik ruht (Ziel: keine unntige ADC-Strung)
  RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
  RCC->APB2ENR |= RCC_APB2ENR_IOPBEN;
  RCC->APB2ENR |= RCC_APB2ENR_IOPCEN;
  RCC->APB2ENR |= RCC_APB2ENR_IOPDEN;
  GPIOA->CRL = 0x88888888;
  GPIOA->CRH = 0x88888888;
  GPIOB->CRL = 0x88888888;
  GPIOB->CRH = 0x88888888;
  GPIOC->CRL = 0x88888888;
  GPIOC->CRH = 0x88888888;
  GPIOD->CRL = 0x88888888;
  GPIOD->CRH = 0x88888888;
  GPIOA->ODR = 0xffff9fff;
  GPIOB->ODR = 0xffffffff;
  GPIOC->ODR = 0xffffffff;
  GPIOD->ODR = 0xffffffff;

  // PWM an PB6 (normal) und PB7 (invertiert)
  {
    RCC->APB1ENR |= RCC_APB1ENR_TIM4EN;
    TIM4->ARR  = ADCSamplesPerPeriod-1;
    TIM4->CCR1 = ADCSamplesPerPeriod/2;
    TIM4->CCR2 = ADCSamplesPerPeriod/2;
    TIM4->SMCR |= TIM_SMCR_TS_0*1;      // 1: Slave Trigger "Internal Trigger 1 (ITR1)" -> TIM2
    TIM4->SMCR |= TIM_SMCR_SMS_0*7;     // 7: Slave Trigger "Selected trigger clock the counter"
    TIM4->CCER |= TIM_CCER_CC1E;
    TIM4->CCER |= TIM_CCER_CC2E;
    TIM4->CCER |= TIM_CCER_CC2P;
    TIM4->CCMR1 |= TIM_CCMR1_OC1M_0*7; //  PWM mode 2 - inactive as long as TIMx_CNT<TIMx_CCR1
    TIM4->CCMR1 |= TIM_CCMR1_OC2M_0*7; //  PWM mode 2 - inactive as long as TIMx_CNT<TIMx_CCR1
    TIM4->CR1 = TIM_CR1_CEN;
    // PB6 und PB7 auf Alternate function; Open-drain output; 2 MHz
    RCC->APB2ENR |= RCC_APB2ENR_IOPBEN;
    GPIOB->CRL = (GPIOB->CRL&0xf0ffffff)|0x0e000000;
    GPIOB->CRL = (GPIOB->CRL&0x0fffffff)|0xe0000000;
  }

  // PA7: Analog
  RCC->APB2ENR |= RCC_APB2ENR_IOPAEN;
  GPIOA->CRL = (GPIOA->CRL&0x0fffffff)|0x00000000;

  // ADC_Frequenz einstellen (TimedAdc.TimArr[] definieren)
  TimedAdc_Init(&TimedAdc, 0, 77500);

  // TIM2
  RCC->APB1ENR |= RCC_APB1ENR_TIM2EN;
  TIM2->ARR = TimedAdc.TimArr[0]; // Achtung: wird per DMA1_Channel2 manipuliert!
  TIM2->CCR1 = 10;                         // TIM3-Zhler-Trigger
  TIM2->CCR2 = 10;                         // ADC-Start-Trigger
  TIM2->CCER = TIM_CCER_CC2E;              // CC2 output enabled (fr ADC-Start-Trigger)
  TIM2->CCMR1 = TIM_CCMR1_OC2M_0*7;        // 7: PWM mode 2 - active TIMx_CNT>=TIMx_CCR1 (fr ADC-Start-Trigger)
  TIM2->CR2 = TIM_CR2_MMS_0*3;             // 3: Master Trigger Output "Compare CC1IF" (fr TIM3-Trigger)
  TIM2->DIER = TIM_DIER_UDE;               // Update Event (fr DMA1_Channel2-Trigger)

  // Init ADC
  // my_assert(fMHz*1000*1000>32*fCarrier*ADCSamplesPerPeriod, _ErrAdcSampleTime); // ADC braucht mind. 32 CPU-Takte
  RCC->CFGR   |= RCC_CFGR_ADCPRE_0*0; // 0 -> PCLK2/2
  RCC->APB2ENR |= RCC_APB2ENR_ADC1EN;
  {
    // Calibrate ADC
    ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
    ADC1->CR2   |= ADC_CR2_ADON;        // start (single) conversion (1->1)
    while((ADC1->SR&ADC_SR_EOC)==0);    // wait for the end of conversion
    ADC1->CR2   |= ADC_CR2_RSTCAL;      // reset calibration
    while((ADC1->CR2&ADC_CR2_RSTCAL));  // wait for reset
    ADC1->CR2   |= ADC_CR2_CAL;         // start calibration
    while((ADC1->CR2&ADC_CR2_CAL)!=0);  // wait till calibration is done
    ADC1->CR2   &= ~ADC_CR2_ADON;       // ADC OFF
  }
  ADC1->CR2   |= ADC_CR2_DMA;         // DMA Enabled
  ADC1->CR2   |= ADC_CR2_EXTTRIG;     // External trigger
  ADC1->CR2   |= ADC_CR2_EXTSEL_0*3;  // 3 -> Trigger TIM2_CC2
  ADC1->SQR1  |= ADC_SQR1_L_0*(1-1);  // 0 -> 1 Kanal
  ADC1->SQR3  |= ADC_SQR3_SQ1_0*7;    // 7 -> PA7
  ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*0;  // 0 -> 14 Takte (T0+Tx=12.5+1.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*1;  // 1 -> 20 Takte (T0+Tx=12.5+7.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*2;  // 2 -> 26 Takte (T0+Tx=12.5+13.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*3;  // 3 -> 41 Takte (T0+Tx=12.5+28.5) fr PA7
  // ADC1->SMPR2 |= ADC_SMPR2_SMP7_0*4;  // 4 -> 54 Takte (T0+Tx=12.5+41.5) fr PA7
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC ON (0->1)
  ADC1->CR2   |= ADC_CR2_ADON;        // ADC Start (1->1) (->1ExtraSample, damit TIM3 synchron zum Sample)

  // Init DMA1_CH1 zum Lesen der ADC-Ergebnisse
  RCC->AHBENR |= RCC_AHBENR_DMA1EN;
  DMA1_Channel1->CNDTR = AdcBufCnt;
  DMA1_Channel1->CPAR  = (uint32_t)&ADC1->DR;
  DMA1_Channel1->CMAR  = (uint32_t)TimedAdc.ValArr;
  DMA1_Channel1->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *1) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *0) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // Init DMA1_CH2 zum Setzen der ADC-Samplezeitpunktabstnde
  DMA1_Channel2->CNDTR = AdcBufCnt;
  DMA1_Channel2->CPAR  = (uint32_t)&TIM2->ARR;
  DMA1_Channel2->CMAR  = (uint32_t)TimedAdc.TimArr; // Zeitabstnde
  DMA1_Channel2->CCR = (DMA_CCR1_MEM2MEM *0) |
                       (DMA_CCR1_PL_0    *0) |
                       (DMA_CCR1_MSIZE_0 *0) |
                       (DMA_CCR1_PSIZE_0 *1) |
                       (DMA_CCR1_MINC    *1) |
                       (DMA_CCR1_PINC    *0) |
                       (DMA_CCR1_CIRC    *1) |
                       (DMA_CCR1_DIR     *1) |
                       (DMA_CCR1_TEIE    *0) |
                       (DMA_CCR1_HTIE    *0) |
                       (DMA_CCR1_TCIE    *0) |
                       (DMA_CCR1_EN      *1);

  // alles in Gang setzen
  TIM2->CR1 = TIM_CR1_CEN;

  // verwendete Spulenparameter anzeigen
  uprintf("Sendespulenparameter: Coil_d_mm=%i; Coil_N_1=%i; Coil_Resistor_Ohm=%i\r\n",
          Coil_d_mm,
          Coil_N_1,
          Coil_Resistor_Ohm);

  // Messen
  while(true)
  {
    int iMax = 0;
    // ### gesamten Frequenzbereich nach AmpArr[]-Index vom Maximum (iMax) absuchen
    {
      int f = MinFrequency;
      int Sum = 0;
      for(int i=0; i<ScanFreqCnt; i++)
      {
        int nPeriod = Scan_Cycle_ms*f/ScanFreqCnt/1000;
        int A = Measure(&TimedAdc, f, nPeriod);
        Sum += A;
        if(i>=ScanAvrCnt)
        {
          Sum -= AmpArr[i-ScanAvrCnt];
          AmpArr[i-ScanAvrCnt] = Sum;
          if(AmpArr[i-ScanAvrCnt]>AmpArr[iMax]) iMax=i-ScanAvrCnt;
        }
        AmpArr[i] = A;
        f += f/fStep;
      }
    }
    // ### Bandbreite (df*2) und Resonanzfrequenz (f1) zu iMax bestimmen
    int f1 = 0;
    int df = 0;
    {
      int f0 = 0;
      int f2 = 0;
      int f = MinFrequency;
      int Amp0 = 70*AmpArr[iMax]/100; // "70/100" =  -3db
      for(int i=-ScanAvrCnt/2; i<ScanFreqCnt-ScanAvrCnt; i++)
      {
        if(i>=0)
        {
          if(f0==0 && AmpArr[i]>Amp0) f0=f;
          if(i==iMax) f1=f;
          if(AmpArr[i]>Amp0) f2=f;
        }
        f += f/fStep;
      }
      df = (f2-f0)/2;
    }
    // ### schrittweise Bandbreite und Resonanzfrequenz genauer bestimmen
    // Dazu erst die Amplitude vom Maximum (f1) und den beiden Grenzfrequenzen
    // (f1+-df) bestimmen. Dann mit Hilfe der Amplitudenfunktion (Nherung:
    // U/U0=1/sqr(4*(df/fb)^2+1)) die Resonanzfrequenz und Bandbandbreite
    // berechnen.
    int nPeriods = ScanAvrCnt*Scan_Cycle_ms/ScanFreqCnt*f1/1000;
    for(int iLoop=0; iLoop<IterationCnt; iLoop++)
    {
      int a0 = Measure(&TimedAdc, f1-df, nPeriods);
      int a1 = Measure(&TimedAdc, f1   , nPeriods);
      int a2 = Measure(&TimedAdc, f1+df, nPeriods);
      int64_t a12 = (int64_t)a1*a1;
      int64_t s0 = a12/a0;
      int64_t s2 = a12/a2;
      s0 = s0*s0-a12;
      s2 = s2*s2-a12;
      if(s0<0) s0=0;
      if(s2<0) s2=0;
      s0 = int64_sqrt(s0);
      s2 = int64_sqrt(s2);
      f1 += (int)(2*df*s0/(s0+s2))-df;
      df = (int)((int64_t)2*df*a1/(s0+s2));
      if(df<10) df=10;
      if(f1<MinFrequency) f1=MinFrequency;
      if(f1>MaxFrequency) f1=MaxFrequency;
      nPeriods *= 2; // erhhen muss man ihn; aber ist *2 nicht etwas viel?
    }
    int Amp = Measure(&TimedAdc, f1   , nPeriods);
    int Amp_Uss_uV = (int)((int64_t)Amp*STM32_VCC_mV*1000*4/ADCRange/256); // Amplitude Spitze-Spitze

    // Berechnung der konkreten Feldstrke der Sendespule "E_uV_m" (Feldstrke in Mikrovolt pro Meter)
    // - Magnetische Flussdichte im Mittelpunkt einer Leiterschleife (Einheit T = kg/A/s/s = V*s/m/m)
    //   B = u0*N*I/d
    // - Elektrische Feldstrke (V/m)
    //   E = B*c
    // - Feldkonstanten
    //   - Magnetische Feldkonstante
    //     u0 = 1.2566e-6 N/A^2
    //   - Lichgeschwindigkeit
    //     c = 3e8 m/s
    // - Gleichungen
    //   const double Is_Rectangle = 0.001*STM32_VCC_mV/Coil_Resistor_Ohm; // Spitzenwert vom Sendespulenstrom
    //   const double Is_Sinus = Is_Rectangle*4/M_PI;                      // Spitzenwert des Sinusanteils vom Sendespulenstrom
    //   const double Ieff = Is_Sinus/1.4142136;                           // Effektivstrom in der Sendespule fr die Messfrequenz
    //   const double E = 1.2566e-6*Coil_N_1*Ieff/(0.001*Coil_d_mm)*3e8;   // resultierende Feldstrke (E=0*N*I/d/c)
    //   const double E_uV_m = E*1e6;                                      // eine auf den Integerwertebereich skalierte Konstante
    //   nun alles eingesetzt...
    static const int E_uV_m = (int)(1e6*1.2566e-6*Coil_N_1*(0.001*STM32_VCC_mV/Coil_Resistor_Ohm*4/M_PI/1.4142136)/(0.001*Coil_d_mm)*3e8); // E = 1e6*0*N*I/d/c

    // am ADC anliegende Eingangsspannung
    int UIn_eff_uV = 1000/2*Amp_Uss_uV/(int)(1000*sqrt(2)); // "Ueff=Uss/2/sqr(2)"

    // kSystem_mV_Vm = "Empfindlichkeit des Schwingkreises inkl. Verstrker in mV/(V/m)"
    int kSystem_mV_Vm = 1000*UIn_eff_uV/E_uV_m;

    while(_Error!=_ErrOk)
    {
      uprintf("Error #%i\r\n", _Error);
    }

    uprintf("f =%6i Hz;   fb =%5i Hz;   Uv/E =%5.3i V/Vm;   UssADC=%4i mV",
            f1,
            2*df,
            kSystem_mV_Vm,
            Amp_Uss_uV/1000);
    if(Amp_Uss_uV< 100*1000) uprintf("; UssADC ist zu niedrig!");
    if(Amp_Uss_uV>1000*1000) uprintf("; UssADC ist zu hoch!");
    uprintf("\r\n");
  }
  return 0;
}
